@@ -13,15 +13,17 @@ Test::Group - Group together related tests in a test suite
 
 =head1 VERSION
 
-Test::Group version 0.15
+Test::Group version 0.16
 
 =cut
 
 use vars qw($VERSION);
-$VERSION = '0.15';
+$VERSION = '0.16';
 
 =head1 SYNOPSIS
 
+Basics:
+
 =for tests "synopsis-success" begin
 
     use Test::More no_plan => 1;
@@ -36,21 +38,47 @@ $VERSION = '0.15';
 
 =for tests "synopsis-success" end
 
+Failed subtests are displayed I<before> the result of the test they belong to.
+For instance,
+
 =for tests "synopsis-fail" begin
 
+    use Test::More no_plan => 1;
+    use Test::Group;
+
     test "this test group will fail", sub {
         ok 1, "sub test blah";
-        is "foo", "bar";                    # Oops!
+        is "foo", "bar", "I so wish they were the same...";
         ok 1;
         like   "blah blah blah", qr/bla/;
     };
 
+=for tests "synopsis-fail" end
+
+produces something like this:
+
+    #   Failed test 'I so wish they were the same...'
+    #   in test.pl at line 6.
+    #          got: 'foo'
+    #     expected: 'bar'
+    not ok 1 - this test group will fail
+    #   Failed test 'this test group will fail'
+    #   in test.pl at line 9.
+    1..1
+
+Exceptions in tests are not fatal:
+
+=for tests "synopsis-die" begin
+
     test "this test will fail but the suite will proceed", sub {
         pass;
         die;
     };
 
-=for tests "synopsis-fail" end
+
+=for tests "synopsis-die" end
+
+Test::More style TODO support:
 
 =for tests "synopsis-TODO" begin
 
@@ -69,6 +97,8 @@ $VERSION = '0.15';
 
 =for tests "synopsis-TODO" end
 
+Misc:
+
 =for tests "synopsis-misc" begin
 
     # Don't catch exceptions raised in test groups later on
@@ -181,7 +211,9 @@ predicates is straightforward with I<Test::Group>.  For example,
     sub foobar_ok {
         my ($text, $name) = @_;
         $name ||= "foobar_ok";
+        local $Test::Group::Level = $Test::Group::Level + 1;
         test $name => sub {
+           local $Test::Group::InPredicate = 1;
            like($text, qr/foo/, "foo ok");
            like($text, qr/bar/, "bar ok");
         };
@@ -196,6 +228,14 @@ predicate (in particular, it always counts for a single test, honors
 L<Test::More/TODO: BLOCK> constructs, etc); and of course, users of
 I<Test::Group> can freely call I<foobar_ok> from within a group.
 
+Adding 1 to C<$Test::Group::Level> causes the location of the call to
+foobar_ok() to be shown if a test fails, rather than the location of
+the test() call within foobar_ok().
+
+Setting C<$Test::Group::InPredicate> to a true value prevents the
+location of individual failing subtests within test groups from being
+shown.
+
 =head2 TODO Tests
 
 As shown in L</SYNOPSIS>, L<Test::More>'s concept of TODO tests is
@@ -253,6 +293,19 @@ one to be found in L</SYNOPSIS>) will succeed as expected:
 
 =for tests "TODO gotcha 2" end
 
+=head2 OUTPUT FORMAT
+
+As seen briefly in L</SYNOPSIS>, only top-level test groups (and toplevel
+Test::More tests if any) produce a single "ok" or "not ok" summary line.  Failed
+sub-tests produce non-scoring comment messages (prefixed with "#"); successful
+sub-tests are silent.  This is different from, and predates, the
+L<Test::More/subtest> functionality.
+
+=head2 PLUGIN INTERFACE
+
+A simple plugin interface allows module authors to write extensions
+to I<Test::Group>.  See L<Test::Group::Extending> for details.
+
 =cut
 
 use 5.004;
@@ -264,7 +317,7 @@ BEGIN { die "Need Test::Simple version 0.59 or later, sorry"
 use IO::File;
 use File::Spec;
 
-my $classstate_verbose;
+my $classstate_verbose = $ENV{PERL_TEST_GROUP_VERBOSE};
 my $classstate_skipcounter;
 my $classstate_skipreason;
 my $classstate_testonly_reason;
@@ -272,6 +325,10 @@ my $classstate_testonly_criteria = sub { 1 };
 my $classstate_catchexceptions = 1;
 my $classstate_logfile;
 my $classstate_logfd;
+my @classstate_plugins;
+
+our $Level = 0;
+our $InPredicate;
 
 =head2 FUNCTIONS
 
@@ -281,11 +338,12 @@ script. They are all exported by default.
 =cut
 
 use Exporter;
-use vars qw(@ISA @EXPORT);
-@ISA    = qw(Exporter);
-@EXPORT = qw(test skip_next_test skip_next_tests
-             begin_skipping_tests end_skipping_tests
-             test_only);
+use vars qw(@ISA @EXPORT @EXPORT_OK);
+@ISA       = qw(Exporter);
+@EXPORT    = qw(test skip_next_test skip_next_tests
+                begin_skipping_tests end_skipping_tests
+                test_only);
+@EXPORT_OK = qw(next_test_plugin);
 
 =head3 test ($name, $groupsub)
 
@@ -392,6 +450,7 @@ MESSAGE
     # TODO status across to Test::Builder; the trick has an adherence
     # in L</ok>, which see.
     local *Test::Builder::todo = sub { $TODO_string };
+    local $Test::Builder::Level = $Test::Builder::Level + $Level;
     $Test->ok($OK, $name);
     return $OK ? 1 : 0;
 }
@@ -482,6 +541,24 @@ sub test_only (;$$) {
     }
 }
 
+=head2 PLUGIN FUNCTIONS
+
+The following function relates to the plugin interface. It is not
+exported be default.  See L<Test::Group::Extending> for details.
+
+=head3 next_test_plugin ( PLUGIN )
+
+Installs a plugin for the next test group. PLUGIN must be a
+subroutine reference.
+
+=cut
+
+sub next_test_plugin (&) {
+    my $plugin = shift;
+
+    push @classstate_plugins, $plugin;
+}
+
 =head1 CLASS METHODS
 
 A handful of class methods are available to tweak the behavior of this
@@ -501,6 +578,9 @@ subtest within top-level test groups. At level 3, the subtests of test
 groups nested within top level test groups also get diagnostic lines,
 and so on.
 
+The default verbosity level is 0, or the value of the
+C<PERL_TEST_GROUP_VERBOSE> environment variable if it is set.
+
 =cut
 
 sub verbose { shift; $classstate_verbose = shift }
@@ -610,6 +690,9 @@ into L</subtests>.  Invoking C<< ->new($name, $sub) >> then C<<
 except that I<test()> additionally passes along the test group results
 to L<Test::Builder>.
 
+If any plugins have been set, they are applied to the test group and
+the list of plugins is cleared.
+
 =cut
 
 sub run {
@@ -628,11 +711,17 @@ sub run {
     Test::Builder->new->diag("Running group of tests - $self->{name}")
         if ($classstate_verbose);
 
+    my $code = $self->{code};
+    if (my @plugins = @classstate_plugins) {
+        $code = sub { $self->_run_code_via_plugins(@plugins) };
+        @classstate_plugins = ();
+    }
+
     my ($exn);
 
     $self->_hijack();    # BEGIN CRITICAL SECTION
     my $exception_raised = !
-        $self->_run_with_local_TODO($self->{callerpackage}, $self->{code});
+        $self->_run_with_local_TODO($self->{callerpackage}, $code);
     $self->_unhijack();  # END CRITICAL SECTION
 
     if ($exception_raised) {
@@ -758,7 +847,9 @@ sub ok {
 
 	if( defined $testname ) {
 	    $T->diag(qq[  $msg test '$testname'\n]);
-	    $T->diag(qq[  in $file at line $line.\n]);
+	    unless ($InPredicate) {
+		$T->diag(qq[  in $file at line $line.\n]);
+	    }
 	} else {
 	    $T->diag(qq[  $msg test in $file at line $line.\n]);
 	}
@@ -1074,6 +1165,28 @@ sub _run_with_local_TODO {
     return $retval;
 }
 
+=head3 _run_code_via_plugins (@plugins)
+
+Invokes the subroutine reference stored as C<$self->{code}> via
+the chain of plugins in C<@plugins>.
+
+=cut
+
+sub _run_code_via_plugins {
+    my ($self, $plugin, @more) = @_;
+
+    if ($plugin) {
+        my $old_inp = $InPredicate;
+        local $InPredicate = 1;
+        $plugin->(sub{
+            local $InPredicate = $old_inp;
+            $self->_run_code_via_plugins(@more);
+        });
+    } else {
+        $self->{code}->();
+    }
+}
+
 =head3 _skip ($reason)
 
 Private setter called from L</run> when the test sub is not to be
